! A simple drawing program
! Copyright (C) 2003 Nils M Holm

module draw(t3x, char, string, iostream, ttyctl, gfxctl, util);

object	t[t3x], chr[char], str[string], tty[ttyctl], gfx[gfxctl];
object	u[util];

const	POOLSIZE =	16384;
const	BUFLEN =	1024;
const	STRLEN =	30;

var	Pool[POOLSIZE], Pb, Pt;
var	Fs[2];
var	Scale;
var	Filename::STRLEN;
var	Modified;

clrmsg() gfx.box(-GFXCTL.LIMIT, GFXCTL.LIMIT-Fs[1]*2, GFXCTL.LIMIT,
		GFXCTL.LIMIT, GFXCTL.OFF);

msg(s) do
	clrmsg();
	gfx.puts(-str.length(s)/2*Fs[0], GFXCTL.LIMIT-Fs[1], Scale,
		GFXCTL.ON, s);
end

drawobj(x, y, p, m) do var j, bx, by;
	j := 0;
	while (p[j] \= GFXCTL.LIMIT+1) do
		bx := x;
		by := y;
		x := bx + p[j];
		y := by + p[j+1];
		gfx.line(bx, by, x, y, m);
		j := j+2;
	end
end

redraw() do var i, p;
	gfx.clear(GFXCTL.OFF);
	for (i=Pt, POOLSIZE) do
		p := Pool[i];
		drawobj(p[0], p[1], @p[2], GFXCTL.ON);
	end
	msg("Left = draw, Right = menu");
end

alloc(n, bot) do var p;
	if (Pb + n >= Pt) do
		msg("Out of memory");
		return 0;
	end
	ie (bot) do
		p := @Pool[Pb];
		Pb := Pb + n;
	end
	else do
		Pt := Pt - n;
		p := @Pool[Pt];
	end
	return p;
end

sizeof(p) do var n;
	n := 0;
	while (p[n] \= GFXCTL.LIMIT+1) n := n+1;
	return n+1;
end

frame(x, y, dx, dy, m) do
	gfx.line(x, y, dx, y, m);
	gfx.line(x, y, x, dy, m);
	gfx.line(x, dy, dx, dy, m);
	gfx.line(dx, y, dx, dy, m);
end

frame_on(x, y, dx, dy) frame(x, y, dx, dy, GFXCTL.ON);

release() do var e[GFXCTL.EVENT];
	gfx.poll(e, 0);
	while (e[GFXCTL.EV_BUTTONS]) gfx.poll(e, 1);
end

menu(items) do
	var	x, y;
	var	n, i, h, w, k;
	var	e[GFXCTL.EVENT], att, cx, cy, curr, prev;
	var	v, d;

	gfx.poll(e, 0);
	x := e[GFXCTL.EV_X];
	y := e[GFXCTL.EV_Y];
	w := 0;
	n := 0;
	while (items[n]) do
		k := str.length(items[n]);
		if (k > w) w := k;
		n := n+1;
	end
	d := Fs[1] / 5;
	v := Fs[1] + d;
	h := (n) * v + Fs[1] / 4;
	w := (w+1) * Fs[0];
	if (x+w > GFXCTL.LIMIT) x := GFXCTL.LIMIT-w;
	if (y+h > GFXCTL.LIMIT) y := GFXCTL.LIMIT-h;
	gfx.box(x, y, x+w, y+h, GFXCTL.OFF);
	frame_on(x, y, x+w, y+h);
	for (i=0, n) do
		gfx.puts(x+Fs[0], y+(i+1)*v, Scale, GFXCTL.ON, items[i]);
	end
	curr := -1;
	prev := -1;
	release();
	while (1) do
		att := gfx.poll(e, 0);
		cx := e[GFXCTL.EV_X];
		cy := e[GFXCTL.EV_Y];
		if (att) leave;
		curr := -1;
		for (i=0, n) do
			if (	cx >= x /\ cx <= x+w /\
				cy >= y+i*v /\ cy <= y+(i+1)*v
			)
				curr := i;
		end
		if (curr \= prev) do
			if (prev \= -1) do
				gfx.box(x, y+prev*v+d, x+w, y+(prev+1)*v+d,
					GFXCTL.OFF);
				gfx.puts(x+Fs[0], y+(prev+1)*v, Scale,
					GFXCTL.ON, items[prev]);
				frame_on(x, y, x+w, y+h);
			end
			if (curr \= -1) do
				gfx.box(x, y+curr*v+d, x+w, y+(curr+1)*v+d,
					GFXCTL.ON);
				gfx.puts(x+Fs[0], y+(curr+1)*v, Scale,
					GFXCTL.OFF, items[curr]);
			end
			prev := curr;
		end
	end
	redraw();
	release();
	return curr;
end

confirm(yes, no) return menu([(yes, no, 0)]) = 0;

discard() return \Modified-> 1: confirm("Discard changes", "No, wait");

reset() do
	Pb := 0;
	Pt := POOLSIZE;
	Modified := 0;
	Filename::0 := 0;
	redraw();
end

validchar(c) return chr.alpha(c) \/ chr.digit(c) \/ c = '.' \/ c = '_';

prompt(p, s) do var k, h, x, y, cy, cx, e[GFXCTL.EVENT], key, r;
	y := GFXCTL.LIMIT-Fs[1];
	cy := y + Fs[1]/4;
	cx := Fs[0];
	clrmsg();
	gfx.puts(-GFXCTL.LIMIT+Fs[0], y, Scale, GFXCTL.ON, p);
	k := str.length(s);
	h := k;
	x := -GFXCTL.LIMIT + (str.length(p)+1) * Fs[0];
	r := 1;
	while (1) do
		if (r) do
			gfx.box(x, GFXCTL.LIMIT-Fs[1]*2, GFXCTL.LIMIT,
				GFXCTL.LIMIT, GFXCTL.OFF);
			gfx.puts(x, y, Scale, GFXCTL.ON, s);
			r := 0;
		end
		gfx.box(x+h*Fs[0], cy-Fs[1], x+h*Fs[0]+cx, cy, GFXCTL.INV);
		gfx.poll(e, 1);
		gfx.box(x+h*Fs[0], cy-Fs[1], x+h*Fs[0]+cx, cy, GFXCTL.INV);
		key := e[GFXCTL.EV_KEY];
		ie (key = TTYCTL.K_LEFT \/ key = 'P'-'@') do
			if (h > 0) h := h-1;
		end
		else ie (key = TTYCTL.K_RIGHT \/ key = 'N'-'@') do
			if (h < k) h := h+1;
		end
		else ie (key = '\b') do
			if (h > 0) do
				h := h-1;
				t.memcopy(@s::h, @s::(h+1), k-h+1);
				k := k-1;
				r := 1;
			end
		end
		else ie (key = TTYCTL.K_DEL \/ key = 'D'-'@') do
			if (k > h) do
				t.memcopy(@s::h, @s::(h+1), k-h+1);
				k := k-1;
				r := 1;
			end
		end
		else ie (key = TTYCTL.K_KILL \/ key = 'U'-'@') do
			h := 0;
			k := 0;
			s::0 := 0;
			r := 1;
		end
		else ie (key = '\r') do
			r := k;
			leave;
		end
		else ie (key = TTYCTL.K_ESC) do
			r := -1;
			leave;
		end
		else if (validchar(key) /\ k < STRLEN-1) do
			t.memcopy(@s::(h+1), @s::h, k-h+1);
			s::h := key;
			h := h+1;
			k := k+1;
			r := 1;
		end
	end
	clrmsg();
	return r;
end

save() do
	object	file[iostream];
	var	buf::BUFLEN, nbuf::10;
	var	i, p, j, x, k;

	if (prompt("Save: ", Filename) < 0) return;
	if (file.open(Filename, buf, BUFLEN, IOSTREAM.FWRITE) = -1) do
		msg("File could not be created");
		return;
	end
	msg("Saving...");
	file.writes("DRAW\n");
	for (i=Pt, POOLSIZE) do
		p := Pool[i];
		j := 0;
		x := 0;
		while (1) do
			str.format(nbuf, " %d", [(p[j])]);
			k := str.length(nbuf);
			if (k + x > 70) do
				file.wrch('\n');
				x := 0;
			end
			file.writes(nbuf);
			x := x+k;
			if (p[j] = GFXCTL.LIMIT+1) leave;
			j := j+1;
		end
		file.wrch('\n');
	end
	file.close();
	if (file.error()) do
		msg("File write error");
		return;
	end
	Modified := 0;
	msg("Saved.");
end

getnum(file, p) do var c, v, g;
	while (1) do
		c := send(file, iostream, rdch());
		if (\chr.space(c)) leave;
	end
	if (send(file, iostream, eof())) return -1;
	g := 1;
	v := 0;
	if (c = '-') do
		c := send(file, iostream, rdch());
		g := -1;
	end
	while (chr.digit(c)) do
		v := v*10+chr.value(c);
		c := send(file, iostream, rdch());
	end
	if (\chr.digit(c) /\ \chr.space(c) /\ \send(file, iostream, eof()))
		return -2;
	p[0] := v*g;
	return 0;
end

clear() do
	if (\discard()) return 0;
	reset();
	return 1;
end

load() do
	object	file[iostream];
	const	MAX = 80;
	var	buf::BUFLEN, lbuf::MAX;
	var	n, stat, p, obj;
	var	name::STRLEN;

	if (\clear()) return;
	name::0 := 0;
	if (prompt("Load: ", name) < 0) return;
	if (file.open(name, buf, BUFLEN, IOSTREAM.FREAD) = -1) do
		msg("File could not be opened");
		return;
	end
	msg("Loading...");
	file.reads(lbuf, MAX);
	if (str.comp(lbuf, "DRAW\n")) do
		file.close();
		msg("Not a draw file");
		return;
	end
	reset();
	obj := Pool;
	stat := getnum(file, @n);
	while (\stat) do
		p := alloc(1, 1);
		if (p) p[0] := n;
		if (n = GFXCTL.LIMIT+1) do
			p := alloc(1, 0);
			if (p) p[0] := obj;
			obj := @Pool[Pb];
		end
		if (\p) do
			stat := -3;
			leave;
		end
		stat := getnum(file, @n);
	end
	file.close();
	str.copy(Filename, name);
	ie (stat = -2) do
		reset();
		msg("Draw file is corrupt");
	end
	else if (stat = -3) do
		msg("Not enough memory");
		Filename::0 := 0;
	end
	Modified := 0;
	redraw();
end

boundingbox(obj, px0, py0, pxn, pyn) do var cx, cy, x0, y0, xn, yn, i;
	x0 := GFXCTL.LIMIT;
	y0 := GFXCTL.LIMIT;
	xn := -GFXCTL.LIMIT;
	yn := -GFXCTL.LIMIT;
	cx := obj[0];
	cy := obj[1];
	i := 2;
	while (1) do
		if (cx < x0) x0 := cx;
		if (cy < y0) y0 := cy;
		if (cx > xn) xn := cx;
		if (cy > yn) yn := cy;
		if (obj[i] = GFXCTL.LIMIT+1) leave;
		cx := cx + obj[i];
		cy := cy + obj[i+1];
		i := i+2;
	end
	px0[0] := x0;
	pxn[0] := xn;
	py0[0] := y0;
	pyn[0] := yn;
end

inobject(obj, x, y) do var x0, y0, xn, yn, i;
	boundingbox(obj, @x0, @y0, @xn, @yn);
	return x0 <= x /\ x <= xn /\ y0 <= y /\ y <= yn;
end

select(p) do var i, e[GFXCTL.EVENT], r, x, y;
	msg("Left = select, Right = cancel");
	gfx.poll(e, 1);
	release();
	if (e[GFXCTL.EV_BUTTONS] & GFXCTL.B_RIGHT) return 0;
	x := e[GFXCTL.EV_X];
	y := e[GFXCTL.EV_Y];
	for (i=Pt, POOLSIZE) do
		if (inobject(Pool[i], x, y)) do
			p[0] := Pool[i];
			return i;
		end
	end
	return 0;
end

destframe(p, x, y) drawobj(x, y, @p[2], GFXCTL.INV);

destination(what, p, pnx, pny) do
	var	ox, oy;
	var	x, y, lx, ly;
	var	e[GFXCTL.EVENT];
	var	b::STRLEN;

	msg(str.format(b, "Left = %s, Right = cancel", [(what)]));
	gfx.poll(e, 0);
	x := e[GFXCTL.EV_X];
	y := e[GFXCTL.EV_Y];
	lx := -1;
	while (1) do
		if (lx \= x \/ ly \= y) do
			if (lx \= -1) destframe(p, lx, ly);
			destframe(p, x, y);
		end
		lx := x;
		ly := y;
		gfx.poll(e, 0);
		x := e[GFXCTL.EV_X];
		y := e[GFXCTL.EV_Y];
		if (e[GFXCTL.EV_BUTTONS]) leave;
	end
	release();
	destframe(p, x, y);
	if (e[GFXCTL.EV_BUTTONS] & GFXCTL.B_RIGHT) return 0;
	pnx[0] := x;
	pny[0] := y;
	return 1;
end

move() do var p, x, y;
	if (\select(@p)) return;
	if (\destination("move", p, @x, @y)) return;
	Modified := 1;
	p[0] := x;
	p[1] := y;
	redraw();
end

copy() do var p, q, o, n, x, y;
	if (\select(@p)) return;
	if (\destination("copy", p, @x, @y)) return;
	Modified := 1;
	n := sizeof(p);
	q := alloc(n, 1);
	if (q) do
		o := alloc(1, 0);
		if (o) do
			o[0] := q;
			t.memcopy(q, p, n * t.bpw());
			q[0] := x;
			q[1] := y;
		end
	end
	redraw();
end

delete() do
	var	p, x0, y0, xn, yn;
	var	e[GFXCTL.EVENT];
	var	i, j, x, n;

	if (\select(@p)) return;
	boundingbox(p, @x0, @y0, @xn, @yn);
	frame(x0, y0, xn, yn, GFXCTL.INV);
	msg("Left = delete, Right = cancel");
	gfx.poll(e, 1);
	frame(x0, y0, xn, yn, GFXCTL.INV);
	if (\(e[GFXCTL.EV_BUTTONS] & GFXCTL.B_LEFT)) return;
	Modified := 1;
	x := POOLSIZE;
	j := Pb;
	for (i=Pt, POOLSIZE) do
		ie (p = Pool[i]) do
			x := i;
		end
		else do
			n := sizeof(Pool[i]);
			if (@Pool[j] \= Pool[i]) do
				t.memcopy(@Pool[j], Pool[i], n*t.bpw());
				Pool[i] := @Pool[j];
			end
			j := j + n;
		end
		Pb := j;
	end
	if (x = POOLSIZE) return;
	if (x-Pt) t.memcopy(@Pool[Pt+1], @Pool[Pt], (x-Pt) * t.bpw());
	Pt := Pt+1;
	redraw();
end

mainmenu() do var r, x;
	r := menu([
		"Move",
		"Copy",
		"Delete",
		"Load",
		"Save",
		"Clear",
		"Quit",
	0 ]);
	x := 0;
	ie (r = 0) move();
	else ie (r = 1) copy();
	else ie (r = 2) delete();
	else ie (r = 3) load();
	else ie (r = 4) save();
	else ie (r = 5) clear();
	else if (r = 6) x := discard();
	redraw();
	return x;
end

draw() do
	var	e[GFXCTL.EVENT];
	var	x, y, lastx, lasty;
	var	state, pendown, att, p;

	Modified := 1;
	pendown := 1;
	state := 0;
	while (pendown) do
		gfx.wait(10);
		att := gfx.poll(e, 0);
		x := e[GFXCTL.EV_X];
		y := e[GFXCTL.EV_Y];
		if (att) pendown := e[GFXCTL.EV_BUTTONS] & GFXCTL.B_LEFT;
		ie (pendown) do
			ie (state = 0) do
				p := alloc(1, 0);
				if (\p) loop;
				p[0] := @Pool[Pb];
			end
			else if (lastx = x /\ lasty = y) do
				loop;
			end
			p := alloc(2, 1);
			if (\p) loop;
			ie (state) do
				gfx.line(lastx, lasty, x, y, GFXCTL.ON);
				p[0] := x-lastx;
				p[1] := y-lasty;
			end
			else do
				p[0] := x;
				p[1] := y;
			end
			state := pendown;
		end
		else if (state) do
			p := alloc(1, 1);
			if (\p) loop;
			p[0] := GFXCTL.LIMIT+1;
			state := 0;
		end
		lastx := x;
		lasty := y;
	end
end

evloop() do var e[GFXCTL.EVENT], att;
	release();
	while (1) do
		att := gfx.poll(e, 1);
		if (gfx.reshaped()) redraw();
		if (att) do
			ie (e[GFXCTL.EV_BUTTONS] & GFXCTL.B_LEFT)
				draw();
			else if (e[GFXCTL.EV_BUTTONS] & GFXCTL.B_RIGHT)
				if (mainmenu()) leave;
		end
	end
end

w(s) do var nl::4;
	t.write(T3X.SYSERR, s, str.length(s));
	t.write(T3X.SYSERR, t.newline(nl), str.length(nl));
end

init() do
	if (gfx.init("Draw")) do
		w("Graphics init failed.");
		halt 1;
	end
	Scale := 2;
	gfx.fontsize(Scale, Fs);
	reset();
end

fini() do
	gfx.fini();
end

do
	init();
	evloop();
	fini();
end
