! Graphics support for T3X programs.
! Portable routines.
! Copyright (C) 1997,2003 Nils M. Holm.
! See the file LICENSE for conditions of use.

module gfxctl(t3x, string, util);

object	t[t3x], str[string], u[util];

public iclass gfxctl("gfxctl")

 public const	OK	= 0,
		NOGFX	= -1,
		NOMEM	= -2,
		NOFILE	= -3,
		BADFILE	= -4;

 public const	LIMIT	= 16000;

 public const	OFF	= 0,
		ON	= 1,
		INV	= 2;

 public struct	EVENT =
		EV_X,		! coordinates of event
		EV_Y,
		EV_BUTTONS,	! button flags, see below
		EV_KEY;		! keycode or 0

 public const	B_LEFT 		= 1,
		B_MIDDLE	= 2,
		B_RIGHT		= 4;

 const	FSX		= 416,
	FSY		= 833;

 const	POOLSIZE	= 1536;

 var	Vfont;
 var	Pool::POOLSIZE;
 var	Pt;
 var	Vfx, Vfy;

 idecl reshaped(0,0);
 idecl _init(1,1);
 idecl fini(0,0);
 idecl sync(0,0);
 idecl point(3,0);
 idecl line(5,0);
 idecl box(5,0);
 idecl clear(1,0);
 idecl ratio(0,0);
 idecl wait(1,0);
 idecl poll(2,1);

 error(s) do var nl::4;
	self.fini();
	t.write(T3X.SYSERR, "GFX: ", 5);
	t.write(T3X.SYSERR, s, str.length(s));
	t.write(T3X.SYSERR, t.newline(nl), str.length(nl));
	halt -1;
 end

 allocate(size) do var a;
	if (size + Pt >= POOLSIZE) error("Out of memory");
	a := @Pool::Pt;
	Pt := Pt+size;
	return a;
 end

 public loadvf(file, statusp) do var infd, buf::4, ch::1, n, i, k, p;
	Pt := 0;
	Vfont := allocate(128*t.bpw());
	if (Vfont = 0) do
		statusp[0] := NOMEM;
		return 0;
	end
	infd := t.open(file, T3X.OREAD);
	if (infd < 0) do
		statusp[0] := NOFILE;
		Pt := 0;
		return 0;
	end
	if (	t.read(infd, buf, 4) \= 4 \/
		t.memcomp(buf, "VF10", 4)
	) do
		t.close(infd);
		Pt := 0;
		statusp[0] := BADFILE;
		return 0;
	end
	for (i=0, 32) Vfont[i] := 0;
	Vfont[0] := allocate(3);
	if (Vfont[0] = 0) do
		t.close(infd);
		Pt := 0;
		statusp[0] := NOMEM;
		return 0;
	end
	Vfont[0]::0 := 0;
	t.read(infd, ch, 1);
	Vfx := ch::0;
	Vfont[0]::1 := Vfx;
	k := t.read(infd, ch, 1);
	Vfy := ch::0;
	Vfont[0]::2 := Vfy;
	while (k > 0 /\ i < 128) do
		t.read(infd, ch, 1);
		n := ch::0;
		Vfont[i] := allocate(n+1);
		if (Vfont[i] = 0) do
			t.close(infd);
			Pt := 0;
			statusp[0] := NOMEM;
			return 0;
		end
		Vfont[i]::0 := n;
		k := n-> t.read(infd, @Vfont[i]::1, n): 1;
		i := i+1;
	end
	t.close(infd);
	if (i<128) do
		Pt := 0;
		statusp[0] := BADFILE;
		return 0;
	end
	statusp[0] := OK;
	return Vfont;
 end

 public fontsize(scale, fxy) do
	fxy[0] := FSX * scale;
	fxy[1] := FSY * scale;
 end

 public copyfont(area, buffer) do var i, j, bt;
	bt := 0;
	for (i=0, 128) do
		if (area /\ buffer) do
			area[i] := bt;
			ie (\Vfont[i]) do
				buffer[bt] := 0;
				j := 0;
			end
			else do
				for (j=0, Vfont[i]::0)
					buffer[bt+j] := Vfont[i]::j;
			end
		end
		bt := bt + j+1;
	end
	return bt;
 end

 public putv(x, y, scale, mode, v) do var dx, dy, nx, ny, i, cx, cy, d;
	cx := FSX / Vfx * scale;
	cy := FSY / Vfy * scale;
	for (i=1, v::0+1) do
		d := v::i & 0x70 >> 4;
		ie (d = 0)      do dx :=  0; dy := -1; end
		else ie (d = 1) do dx :=  1; dy := -1; end
		else ie (d = 2) do dx :=  1; dy :=  0; end
		else ie (d = 3) do dx :=  1; dy :=  1; end
		else ie (d = 4) do dx :=  0; dy :=  1; end
		else ie (d = 5) do dx := -1; dy :=  1; end
		else ie (d = 6) do dx := -1; dy :=  0; end
		else            do dx := -1; dy := -1; end
		nx := x + (dx * (v::i & 15) * cx);
		ny := y + (dy * (v::i & 15) * cy);
		if (v::i & 0x80) self.line(x, y, nx, ny, mode);
		x := nx;
		y := ny;
	end
 end

 public putc(x, y, scale, mode, c) do
	if (c & ~127) return -1;
	if (\Vfont[c]) return;
	self.putv(x, y, scale, mode, Vfont[c]);
 end

 public puts(x, y, scale, mode, s) do var cx, i, k;
	cx := FSX * scale;
	k := str.length(s);
	for (i=0, k) do
		self.putv(x, y, scale, mode, Vfont[s::i]);
		x := x + cx;
	end
 end

 public init(title) do var p, fp::256, base::240;
	self._init(title);
	if (t.getenv("T3XDIR", base, 240))
		str.copy(base, "/usr/local/t3x/r8");
	str.format(fp, "%s/lib/system.vf", [(base)]);
	if (self.loadvf("font.vf", @p) = 0) do
		if (self.loadvf(fp, @p) = 0) do
			error(p = NOMEM-> "Not enough memory":
				p = NOFILE-> "Cannot open font file":
				p = BADFILE-> "Invalid font file":
				"Unknown error in gfx.init");
		end
	end
 end

end

do end
