! TIDE -- Editor class
! Copyright (C) 2000,2002 Nils M Holm. All rights reserved.
! See the file LICENSE for conditions of use.

module editor();

public class editor(t3x, string)

 object	t[t3x], str[string];

 var	Buffer, Size;
 var	Gap, Egap, Point;
 var	Line;
 var	Block, Eblock;

 public struct	SIZES = S_LENGTH, S_TOTAL, S_LINE;
 public const	E_SPACE = -1, E_NOMATCH = -2, E_OPEN = -3, E_SIZE = -4,
		E_NOBLK = -5, E_SRCDST = -6, E_IO = -7;

 public reset() do
	Gap := 0;
	Egap := Size;
	Point := 0;
	Line := 0;
	Block := 0;
	Eblock := 0;
 end

 public init(buf, n) do
	Buffer := buf;
	Size := n;
	self.reset();
 end

 movegap(p) do
	ie (Gap < p) do
		t.memcopy(@Buffer::Gap, @Buffer::Egap, p-Gap);
		Egap := Egap + p - Gap;
	end
	else if (Gap > p) do
		Egap := Egap + p - Gap;
		t.memcopy(@Buffer::Egap, @Buffer::p, Gap-p);
	end
	Gap := p;
 end

 normalize() movegap(Size-Egap+Gap);

 pos(n) return n<Gap-> n: n+Egap-Gap;

 public _point() return Point;

 public _line() return Line;

 public _offset() return Point-Line;

 public _block() return Block;

 public _eblock() return Eblock;

 public _length() return Size-Egap+Gap;

 public _char() return Buffer::pos(Point);

 public _size(sv) do var i, n, m, lim;
	normalize();
	n := 0;
	m := 1;
	lim := Size-Egap+Gap;
	for (i=0, lim) do
		if (Buffer::i = '\n') do
			m := m+1;
			if (i < Point) n := n+1;
		end
	end
	sv[S_LENGTH] := lim;
	sv[S_TOTAL] := m;
	sv[S_LINE] := n;
	return lim;
 end

 public goto(p) do
	Point := p;
	for (Line = Buffer::pos(Point) = '\n'-> Point-1: Point, -1, -1)
		if (Buffer::pos(Line) = '\n') leave;
	Line := Line+1;
 end

 public gotoln(n) do var i, lim;
	normalize();
	lim := self._length();
	for (i=0, lim) do
		if (\n) leave;
		if (Buffer::i = '\n') n := n-1;
	end
	if (n) return E_SIZE;
	Point := i;
	Line := i;
 end

 public gotocol(n) do var lim;
	lim := self._length();
	for (Point=Line, lim) do
		if (\n) leave;
		if (Buffer::pos(Point) = '\n') leave;
		n := n-1;
	end
	if (n) return E_SIZE;
 end

 public advance(n) do var m, lim;
	if (Point + n > Size-Egap+Gap) return E_SIZE;
	m := 0;
	lim := Point+n;
	for (Point = Point, lim)
		if (Buffer::pos(Point) = '\n') m := m+1;
	self.goto(Point);
	return m;
 end

 public goback(n) do var m, i, lim;
	if (Point < n) return E_SIZE;
	Point := Point - n;
	m := 0;
	for (i = Point, Point + n)
		if (Buffer::pos(Point) = '\n') m := m+1;
	self.goto(Point);
	return m;
 end

 public next(n) do var i, lim;
	normalize();
	lim := Size-Egap+Gap;
	for (i=Point, lim) do
		if (\n) leave;
		if (Buffer::i = '\n') n := n-1;
	end
	if (n) return E_SIZE;
	self.goto(i);
 end

 public prev(n) do var i;
	normalize();
	if (\Line) return E_SIZE;
	for (i=Line-1, -1, -1) do
		if (Buffer::i = '\n') n := n-1;
		if (\n) leave;
	end
	if (n) return E_SIZE;
	while (i > 0 /\ Buffer::(i-1) \= '\n')
		i := i-1;
	self.goto(i);
 end

 public top() do
	Point := 0;
	Line := 0;
 end

 public bottom() do
	Point := Size-Egap+Gap;
	self.goto(Point);
 end

 public home() Point := Line;

 public eol() do var lim;
	lim := Size-Egap+Gap;
	for (Point=Point, lim)
		if (Buffer::pos(Point) = '\n')
			leave;
	return Point-Line;
 end

 public insertc(k) do
	Eblock := 0;
	if (Gap >= Egap) return E_SPACE;
	movegap(Point);
	Buffer::Gap := k;
	Gap := Gap+1;
 end

 public inserts(s) do var k;
	Eblock := 0;
	k := str.length(s);
	if (Gap+k >= Egap) return E_SPACE;
	movegap(Point);
	t.memcopy(@Buffer::Gap, s, k);
	Gap := Gap+k;
 end

 public replacec(k) do
	Eblock := 0;
	movegap(Point);
	if (Egap >= Size) return E_SPACE;
	Buffer::Egap := k;
 end

 public replaces(s) do var k;
	Eblock := 0;
	movegap(Point);
	k := str.length(s);
	if (Egap+k >= Size) return E_SPACE;
	t.memcopy(@Buffer::Egap, s, k);
 end

 public delete(n) do var i, nl;
	Eblock := 0;
	movegap(Point);
	if (Egap+n > Size) return E_SIZE;
	nl := 0;
	for (i=Egap, Egap+n)
		if (Buffer::i = '\n')
			nl := nl+1;
	Egap := Egap+n;
	return nl;
 end

 public copy(offset, buf, len) do var i, c, lim;
	lim := Size-Egap+Gap;
	i := 0; while (i < len-1 /\ offset+i < lim) do
		c := Buffer::pos(offset+i);
		buf::i := c; i := i+1;
		if (c = '\n') leave;
	end
	buf::i := 0;
	return i;
 end

 public copyln(buf, len) return self.copy(Line, buf, len);

 public find(s, start, lim) do var k, i, c;
	k := str.length(s);
	normalize();
	lim := (lim-> lim: Gap) - k;
	c := s::0;
	for (i=start, lim) do
		if (	Buffer::i = c /\
			\t.memcomp(@Buffer::i, s, k)
		)
			return i;
	end
	return E_NOMATCH;
 end

 public rfind(s, start, lim) do var k, i, c;
	k := str.length(s);
	normalize();
	lim := (lim-> lim: Gap) - k;
	c := s::0;
	start := start-1;
	for (i=lim-1, start, -1) do
		if (	Buffer::i = c /\
			\t.memcomp(@Buffer::i, s, k)
		)
			return i;
	end
	return E_NOMATCH;
 end

 public load(name) do var fd;
	self.init(Buffer, Size);
	fd := t.open(name, T3X.OREAD);
	if (fd < 0) return E_OPEN;
	Gap := t.read(fd, Buffer, Size);
	t.close(fd);
	return Gap;
 end

 public load_dos(name) do
 	var b::512, fd, k, len, j1, j2, c;

	self.init(Buffer, Size);
	fd := t.open(name, T3X.OREAD);
	if (fd < 0) return E_OPEN;
	len := 0;
	while (len < Size) do
		k := t.read(fd, b, 512);
		j1 := 0;
		j2 := 0;
		while (j1 < k /\ len+j2 < Size) do
			c := b::j1;
			j1 := j1+1;
			Buffer::(len+j2) := c;
			if (c \= '\r') j2 := j2+1;
		end
		len := len + j2;
		if (k < 512) leave;
	end
	Gap := len;
	t.close(fd);
	return Gap;
 end

 public save(name) do var fd, k;
	normalize();
	fd := t.open(name, T3X.OWRITE);
	if (fd < 0) return E_OPEN;
	k := t.write(fd, Buffer, Gap);
	t.close(fd);
	return k=Gap-> 0: E_IO;
 end

 public save_dos(name) do
 	var b::512, fd, k, len, lim, i, j1, j2, c;

	normalize();
	fd := t.open(name, T3X.OWRITE);
	if (fd < 0) return E_OPEN;
	len := Gap;
	for (i=0, Gap, 256) do
		j1 := 0;
		j2 := 0;
		lim := len<256-> len: 256;
		while (j1 < lim) do
			c := Buffer::(i+j1);
			j1 := j1+1;
			ie (c = '\n') do
				b::j2 := '\r';
				b::(j2+1) := '\n';
				j2 := j2+2;
			end
			else do
				b::j2 := c;
				j2 := j2+1;
			end
		end
		k := t.write(fd, b, j2);
		if (k \= j2) do
			t.close(fd);
			return E_IO;
		end
		len := len - lim;
	end
	t.close(fd);
	return 0;
 end

 public mark(here) do
	Block := here;
	Eblock := 0;
 end

 public setblock(here) Eblock := here;

 public clearblock() Eblock := 0;

 public readblock(name) do var fd, k;
	movegap(Point);
	fd := t.open(name, T3X.OREAD);
	if (fd < 0) return E_OPEN;
	k := t.read(fd, @Buffer::Gap, Egap-Gap);
	t.close(fd);
	if (k = Egap-Gap \/ k < 0) return E_SPACE;
	Block := Gap;
	Eblock := Block+k;
	Gap := Gap+k;
	return k;
 end

 public appendblock(name) do var fd, k;
	if (Eblock <= Block) return E_NOBLK;
	normalize();
	fd := t.open(name, T3X.ORDWR);
	t.seek(fd, 0, T3X.SEEK_END);
	k := t.write(fd, @Buffer::Block, Eblock-Block);
	t.close(fd);
	return k = Eblock-Block-> 0: E_IO;
 end

 public writeblock(name) do var fd, k;
	if (Eblock <= Block) return E_NOBLK;
	normalize();
	fd := t.open(name, T3X.OWRITE);
	k := t.write(fd, @Buffer::Block, Eblock-Block);
	t.close(fd);
	return k = Eblock-Block-> 0: E_IO;
 end

 public deleteblock() do
	if (Eblock <= Block) return E_NOBLK;
	self.goto(Block);
	movegap(Block);
	Egap := Egap+Eblock-Block;
	Eblock := 0;
 end

 public copyblock() do var src, z;
	if (Eblock <= Block) return E_NOBLK;
	if (Point > Block /\ Point < Eblock) return E_SRCDST;
	z := Eblock-Block;
	if (z > Egap - Gap) return E_SPACE;
	movegap(Point);
	src := pos(Block);
	t.memcopy(@Buffer::Gap, @Buffer::src, z);
	Gap := Gap+z;
	Block := Gap-z;
	Eblock := Gap;
 end

 public moveblock() do var z, b0, bn, r, n0, nn;
	b0 := Block;
	bn := Eblock;
	z := Eblock-Block;
	if (Point >= Block /\ Point < Eblock) return E_SRCDST;
	r := self.copyblock();
	n0 := Block;
	nn := Eblock;
	if (r < 0) return r;
	ie (Block < b0) do
		Block := b0+z;
		Eblock := bn+z;
	end
	else do
		Block := b0;
		Eblock := bn;
		n0 := n0-z;
		nn := nn-z;
	end
	self.deleteblock();
	Block := n0;
	Eblock := nn;
 end

 public deleteln() do
	Eblock := 0;
	self.home();
	self.mark(Point);
	if (self.next(1) < 0) self.eol();
	self.setblock(Point);
	self.home();
	return self.deleteblock();
 end

 public findmatch() do var i, n;
	normalize();
	n := 0;
	ie (Buffer::Point = '(') do
		for (i=Point, Gap) do
			ie (Buffer::i = '(') n := n+1;
			else if (Buffer::i = ')') n := n-1;
			if (\n) leave;
		end
	end
	else ie (Buffer::Point = ')') do
		for (i=Point, -1, -1) do
			ie (Buffer::i = ')') n := n+1;
			else if (Buffer::i = '(') n := n-1;
			if (\n) leave;
		end
	end
	else do
		return E_NOMATCH;
	end
	return n-> E_NOMATCH: i;
 end

 public marklist() do var i, j, n;
	normalize();
	Block := 0;
	Eblock := 0;
	n := Buffer::Point = ')'-> 0: 1;
	for (i=Point, -1, -1) do
		ie (Buffer::i = ')') n := n+1;
		else if (Buffer::i = '(') n := n-1;
		if (\n) leave;
	end
	if (n) return E_NOMATCH;
	n := Buffer::Point = '('-> 0: 1;
	for (j=Point, Size) do
		ie (Buffer::j = '(') n := n+1;
		else if (Buffer::j = ')') n := n-1;
		if (\n) leave;
	end
	if (n) return E_NOMATCH;
	Block := i;
	Eblock := j+1;
 end

end

do end
