! TIDE -- File selection class
! Copyright (C) 2000,2002 Nils M Holm. All rights reserved.
! See the file LICENSE for conditions of use.

module filebox();

public class filebox(t3x, string, system, ttyctl)

 object	t[t3x], str[string], sys[system], tty[ttyctl];

 public struct	CSET = C_BACKGROUND, C_FRAME, C_TEXT, C_SOTEXT, C_SELECT;
 public struct	FSET = F_UL, F_UR, F_LL, F_LR, F_HL, F_VL;

 const	MINCOLS = 20,
	MINROWS = 6,
	MAXWIDTH = 64,
	MAXPATHL = 128;

 public const	BUFSIZE = MAXPATHL*2;

 var	Rows, Cols;
 var	Colorset, Frameset;
 var	Dflcset, Dflfset;
 var	Xdir, Xfiles, Fwidth, Fheight, Ypos, Suffix;
 var	Ndir, Nfil;
 var	Spaces::MAXWIDTH;
 var	Pathbuf::MAXPATHL+1, Basepath::MAXPATHL+1;

 init() do
	sys.init();
	tty.init();
	tty.mode(1);
	Rows := tty.columns();
	Cols := tty.lines();
	Dflcset := [
		ttyctl.F_BLACK|ttyctl.B_GREY,	! Background
		ttyctl.F_BLACK|ttyctl.B_GREY,	! Frame
		ttyctl.F_GREY|ttyctl.B_BLACK,	! Text
		ttyctl.F_BLACK|ttyctl.B_GREY,	! SO text
		ttyctl.F_BLACK|ttyctl.B_GREY	! Select
	];
	Dflfset := packed[ '\s', '\s', '\s', '\s', '\s', '\s' ];
	sys.getdir(Basepath, MAXPATHL);
 end

 fini() do
	tty.mode(0);
	tty.fini();
	sys.chdir(Basepath);
	sys.fini();
 end

 box(x, y, dx, dy, frame) do var i, sp::MAXWIDTH+1;
	t.memfill(sp, '\s', dx-x);
	sp::(dx-x) := 0;
	tty.color(Colorset[frame->C_BACKGROUND:C_TEXT]);
	for (i=y, dy+1) do
		tty.move(x, i);
		tty.writes(sp);
	end
	if (\frame) return;
	t.memfill(sp, Frameset::F_HL, dx-x);
	sp::0 := Frameset::F_UL;
	sp::(dx-x-1) := Frameset::F_UR;
	tty.color(Colorset[C_FRAME]);
	tty.move(x, y); tty.writes(sp);
	sp::0 := Frameset::F_LL;
	sp::(dx-x-1) := Frameset::F_LR;
	tty.move(x, dy); tty.writes(sp);
	for (i=y+1, dy) do
		tty.move(x, i);
		tty.writec(Frameset::F_VL);
		tty.move(dx-1, i);
		tty.writec(Frameset::F_VL);
	end
 end

 msg(x, y, m) do
	tty.move(x, y);
	tty.color(Colorset[C_FRAME]);
	tty.writes(m);
	tty.writes(" -- [ESC]");
	while(tty.readc() \= ttyctl.K_ESC)
		;
 end

 listfile(x, y, color, file) do var k, c;
	tty.move(x, y);
	tty.color(color);
	tty.writes(Spaces);
	tty.move(x, y);
	k := str.length(file);
	ie (k > Fwidth) do
		c := file::(Fwidth-1);
		file::(Fwidth-1) := 0;
		tty.writes(file);
		file::(Fwidth-1) := c;
		tty.color(Colorset[C_SOTEXT]);
		tty.writec('+');
		tty.color(Colorset[C_TEXT]);
	end
	else do
		tty.writes(file);
	end
	return k;
 end

 list(x, buffer, count) do
	var	sp::MAXWIDTH+1;
	var	k, i;

	t.memfill(Spaces, '\s', Fwidth);
	Spaces::Fwidth := 0;
	for (i=0, Fheight+1) do
		ie (i < count) do
			k := listfile(x, Ypos+i, Colorset[C_TEXT], buffer);
			buffer := @buffer::(k+1);
		end
		else do
			tty.move(x, Ypos+i);
			tty.color(Colorset[C_TEXT]);
			tty.writes(Spaces);
		end
	end
 end

 prev(buf, ptr) do
	ptr := ptr-2;
	while (ptr /\ buf::ptr) ptr := ptr-1;
	if (ptr) ptr := ptr+1;
	return ptr;
 end

 browse(x, buf, count, dirmode) do
	var	red;
	var	pos, off, ptr, top, k;
	var	i;

	red := 1;
	pos := 0;
	off := 0;
	while (1) do
		if (red) do
			top := 0;
			for (i=0, off) top := top+str.length(@buf::top)+1;
			ptr := top;
			for (i=off, pos) ptr := ptr+str.length(@buf::ptr)+1;
			list(x, @buf::top, count-off);
			red := 0;
		end
		listfile(x, Ypos+pos-off, Colorset[C_SELECT], @buf::ptr);
		tty.move(x+Fwidth, Ypos+pos-off);
		k := tty.readc();
		listfile(x, Ypos+pos-off, Colorset[C_TEXT], @buf::ptr);
		ie (k = ttyctl.K_DOWN \/ k = 'N'-'@') do
			if (pos >= count-1) loop;
			pos := pos+1;
			ptr := ptr+str.length(@Buf::ptr)+1;
			if (pos - off > Fheight) do
				off := off+1;
				red := 1;
			end
		end
		else ie (k = ttyctl.K_UP \/ k = 'P'-'@') do
			if (pos < 1) loop;
			pos := pos-1;
			ptr := prev(buf, ptr);
			if (pos < off) do
				off := off-1;
				red := 1;
			end
		end
		else ie (k = ttyctl.K_NEXT \/ k = 'F'-'@') do
			if (count-Fheight-1 = off) loop;
			off := off+Fheight-1;
			ie (off >= count) do
				off := count-Fheight-1;
				if (off < 0) off := 0;
				pos := count-1;
			end
			else do
				pos := off;
			end
			red := 1;
		end
		else ie (k = ttyctl.K_PREV \/ k = 'B'-'@') do
			if (\off) loop;
			off := off-Fheight+1;
			if (off < 0) off := 0;
			pos := off;
			red := 1;
		end
		else ie (k = ttyctl.K_END \/ k = 'E'-'@') do
			if (count-Fheight-1 = off) loop;
			off := count-Fheight-1;
			if (off < 0) off := 0;
			pos := count-1;
			red := 1;
		end
		else ie (k = ttyctl.K_HOME \/ k = 'A'-'@') do
			off := 0;
			pos := 0;
			red := 1;
		end
		else ie (k = '\r' \/ k = ttyctl.K_RIGHT) do
			return @buf::ptr;
		end
		else ie (k = ttyctl.K_LEFT) do
			return dirmode-> buf: 0;
		end
		else ie (k = ttyctl.K_ESC \/ k = ':') do
			return -1;
		end
		else if (k = '\t') do
			return 0;
		end
	end
 end

 showpath(x, y, width, path) do
	var	buf::MAXWIDTH+1;
	var	k, p;

	tty.color(Colorset[C_FRAME]);
	t.memfill(buf, '\s', width);
	buf::width := 0;
	tty.move(x, y);
	tty.writes(buf);
	k := str.length(path);
	ie (k > width)
		p := str.format(buf, "...%S", [(@path::(k-width+3))]);
	else
		p := path;
	tty.move(x, y);
	tty.writes(p);
 end

 insert(name, set, n, llim, ulim, up) do
	var	k, l, i, p;
	var	buf::MAXPATHL+2;

	l := str.length(name);
	if (l >= MAXPATHL) return;
	p := llim;
	for (i=0, n) do
		k := str.length(@set::p);
		if (str.comp(name, @set::p) < 0) leave;
		p := p+k+1;
	end
	str.copy(buf, name);
	ie (up) do
		if (ulim-p) t.memcopy(@set::(p+l+1), @set::p, ulim-p);
	end
	else do
		if (p-llim) t.memcopy(@set::(llim-l-1), @set::llim, p-llim);
		p := p-l-1;
	end
	str.copy(@set::p, buf);
 end

 navigate(path, buffer) do
	var	dp, fp, p;
	var	dd, k, name;
	var	sb[system.STATBUF];
	var	rescan, key;
	var	sl, dir;

	rescan := 1;
	key := 0;
	if (Suffix) sl := str.length(Suffix);
	while (1) do
		if (rescan) do
			dp := 1024;
			fp := 0;
			sys.chdir(path);
			sys.getdir(Pathbuf, MAXPATHL);
			dd := sys.opendir(Pathbuf);
			if (dd = %1) do
				msg(Xdir, Ypos+Fheight+1,
					"invalid path");
				return 0;
			end
			Ndir := 0;
			Nfil := 0;
			while (1) do
				p := @buffer::fp;
				k := sys.readdir(dd, p, dp-fp);
				if (k*2+2 > dp-fp-2) do
					msg(Xdir, Ypos+Fheight+1,
						"too many files");
					leave;
				end
				if (k < 0) leave;
				if (p::0 = '.' /\ p::1 \= '.') loop;
				! ".." must be checked separately, because
				! stat("..", ...) fails under certain
				! conditions, at least when running
				! PCDOS 5.01. See the KNOWNBUGS file in
				! the T3X compiler archive for details.
				ie (p::0 = '.' /\ p::1 = '.' /\ \p::2) do
					dir := 1;
				end
				else do
					if (sys.stat(p, sb) < 0) loop;
					dir := sb[system.ST_MODE] &
						system.FM_ISDIR \= 0;
				end
				ie (dir) do
					insert(p, buffer, Ndir, dp, 1024, 0);
					dp := dp-k-1;
					Ndir := Ndir+1;
				end
				else do
					if (Suffix /\ (sl > k \/
						str.comp(Suffix,
						@buffer::(fp+k-sl)))
					)
						loop;
					insert(p, buffer, Nfil, 0, fp, 1);
					fp := fp+k+1;
					Nfil := Nfil+1;
				end
			end
			sys.closedir(dd);
			if (\key) list(Xdir, @buffer::dp, Ndir);
			rescan := 0;
		end
		showpath(Xdir, Ypos+Fheight+1, Fwidth*2+2, Pathbuf);
		ie (Nfil /\ \key) do
			key := browse(Xfiles, buffer, Nfil, 0);
		end
		else do
			list(Xfiles, buffer, Nfil);
			key := 0;
		end
		ie (key = -1) return 0;
		else ie (\key) ;
		else return key;
		key := browse(Xdir, @buffer::dp, Ndir, 1);
		ie (key = -1) return 0;
		else if (key) do
			name := key;
			ie (\str.comp(name, "..")) do
				k := str.rscan(Pathbuf, '/');
				Pathbuf::(k=0->1: k) := 0;
			end
			else do
				k := str.length(Pathbuf);
				if (k+str.length(name)+1 >= MAXPATHL) do
					msg(Xdir, Ypos+Fheight+1,
						"path too long");
				end
				Pathbuf::k := '/';
				ie (k < 2)	! root
					str.copy(Pathbuf, name);
				else
					str.copy(@Pathbuf::(k+1), name);
			end
			path := Pathbuf;
			rescan := 1;
		end
	end
 end

 public select(x, y, sx, sy, cs, fs, title, path, sfx, fnbuf) do
	var	fw, tl, k;
	var	buffer::1024, file;

	init();
	if (sx < MINCOLS) sx := MINCOLS;
	if (sy < MINROWS) sx := MINROWS;
	if (sx > MAXWIDTH) sx := MAXWIDTH;
	Colorset := cs = 0-> Dflcset: cs;
	Frameset := fs = 0-> Dflfset: fs;
	tl := str.length(title);
	if (tl > sx-2) do
		fini();
		return 0;
	end
	if (sx mod 2) sx := sx-1;
	if (x < 0) x := (Cols-sx)/2-1;
	if (y < 0) y := (Rows-sy)/2-1;
	fw := (sx-6)/2;
	box(x, y, x+sx, y+sy, 1);
	box(x+2, y+2, x+2+fw, y+sy-2, 0);
	box(x+fw+4, y+2, x+4+fw+fw, y+sy-2, 0);
	tty.move(x+(sx-tl)/2, y);
	tty.color(cs[C_FRAME]);
	tty.writes(title);
	Xdir := x+2;
	Xfiles := x+fw+4;
	Ypos := y+2;
	Fwidth := fw;
	Fheight := sy-4;
	Suffix := sfx;
	file := navigate(path, buffer);
	if (file) do
		str.copy(fnbuf, Pathbuf);
		k := str.length(fnbuf);
		fnbuf::k := '/';
		str.copy(@fnbuf::(k>1->k+1:k), file);
	end
	fini();
	return file-> fnbuf: 0;
 end

end

do end
