! TIDE -- Forms class
! Copyright (C) 2000,2002 Nils M Holm. All rights reserved.
! See the file LICENSE for conditions of use.

module form();

public class form(t3x, ttyctl, menu, prompt)

 object	t[t3x], tty[ttyctl];
 object	p[prompt], m[menu];

 const	MAXWIDTH = 128;

 public struct  CSET = C_BACKGROUND, C_FRAME, C_TEXT, C_SOTEXT, C_SELECT;
 public struct  FSET = F_UL, F_UR, F_LL, F_LR, F_HL, F_VL;

 public struct	ITEM = I_TYPE, I_X, I_Y, I_VALUE, I_DATA;
 public struct	KEYS = K_NEXT, K_PREV, K_DONE;

 public const	T_TEXT = 'T', T_INPUT = 'I',  T_MENU = 'M', T_DONE = 'D';

 var	Dflcset, Dfleset, Dflfset, Dflkset;
 var	Colorset, Ecolrset, Frameset;
 var	Keyset;
 var	Spaces::MAXWIDTH+1;
 var	Itemlist;

 init() do
	tty.init();
	tty.mode(1);
        Dflcset := [
                ttyctl.F_BLACK|ttyctl.B_GREY,   ! Background
                ttyctl.F_BLACK|ttyctl.B_GREY,   ! Frame
                ttyctl.F_BLACK|ttyctl.B_GREY,   ! Text
                ttyctl.F_GREY|ttyctl.B_BLACK,   ! SO text
                ttyctl.F_GREY|ttyctl.B_BLACK,   ! Select
        ];
	Dfleset := Dflcset;
        Dflfset := packed[ '\s', '\s', '\s', '\s', '\s', '\s' ];
	Dflkset := [ ttyctl.K_DOWN, ttyctl.K_UP, ttyctl.K_ESC ];
 end

 fini() do
	tty.mode(0);
	tty.fini();
 end

 framebox(x, y, dx, dy) do var i, line::MAXWIDTH;
	tty.color(Colorset[C_BACKGROUND]);
	for (i=y, dy) do
		tty.move(x, i);
		tty.writes(Spaces);
	end
	tty.color(Colorset[C_FRAME]);
	t.memfill(line, Frameset::F_HL, dx-x);
	line::(dx-x) := 0;
	line::0 := Frameset::F_UL;
	line::(dx-x-1) := Frameset::F_UR;
	tty.move(x, y);
	tty.writes(line);
	for (i=y+1, dy-1) do
		tty.move(x, i); tty.writec(Frameset::F_VL);
		tty.move(dx-1, i); tty.writec(Frameset::F_VL);
	end
	line::0 := Frameset::F_LL;
	line::(dx-x-1) := Frameset::F_LR;
	tty.move(x, dy-1);
	tty.writes(line);
 end

 clear(len) do var b::MAXWIDTH+1;
	t.memfill(b, '\s', len);
	b::len := 0;
	tty.writes(b);
 end

 refresh_input(x, y, field) do var b::MAXWIDTH, z;
	tty.color(Colorset[C_SOTEXT]);
	z := field[I_DATA][0];
	clear(z);
	tty.move(x+field[I_X], y+field[I_Y]);
	t.memcopy(b, field[I_VALUE], z);
	b::z := 0;
	tty.writes(b);
 end

 refresh(x, y) do
	var	i, type;

	i := 0; while (Itemlist[i]) do
		type := Itemlist[i][I_TYPE];
		tty.move(x+Itemlist[i][I_X], y+Itemlist[i][I_Y]);
		ie (type = T_TEXT) do
			tty.color(Colorset[C_TEXT]);
			tty.writes(Itemlist[i][I_DATA]);
		end
		else ie (type = T_INPUT) do
			refresh_input(x, y, Itemlist[i]);
		end
		else ie (type = T_MENU) do
			tty.color(Colorset[C_SOTEXT]);
			tty.writes(Itemlist[i][I_DATA]
				[Itemlist[i][I_VALUE][0]][1]);
		end
		else if (type = T_DONE) do
			tty.color(Colorset[C_SOTEXT]);
			tty.writes(Itemlist[i][I_DATA]);
		end
		i := i+1;
	end
 end

 edit(x, y, dx, dy, field) do var xkeys, k, type, v, mx, my, data, i;
	type := field[I_TYPE];
	ie (type = T_INPUT) do
		xkeys := [ 4, '\t', 0, 0, 0 ];
		xkeys[2] := Keyset[K_NEXT];
		xkeys[3] := Keyset[K_PREV];
		xkeys[4] := Keyset[K_DONE];
		k := p.input(x+field[I_X], y+field[I_Y], field[I_DATA][0],
			field[I_DATA][1], Ecolrset, field[I_VALUE],
			xkeys);
		tty.move(x+field[I_X], y+field[I_Y]);
		refresh_input(x, y, field);
	end
	else ie (type = T_MENU) do
		tty.move(x+field[I_X], y+field[I_Y]);
		tty.color(Colorset[C_SELECT]);
		tty.writes(field[I_DATA][field[I_VALUE][0]][1]);
		k := tty.readc();
		if (k = '\r' \/ k = ttyctl.K_RIGHT) do
			mx := x+field[I_X];
			my := y+field[I_Y];
			data := field[I_DATA];
			for (i=0, 99) if (\data[i]) leave;
			my := my - i/2;
			if (my+i+2 > dy) my := dy-i-2;
			if (my < y) my := y;
			v := m.select(mx, my, field[I_VALUE][0], Ecolrset,
				Frameset, data);
			if (v >= 0) field[I_VALUE][0] := v;
		end
		ie (k = '\r' \/ k = ttyctl.K_RIGHT) do
			framebox(x, y, dx, dy);
			refresh(x, y);
		end
		else do
			tty.move(x+field[I_X], y+field[I_Y]);
			tty.color(Colorset[C_SOTEXT]);
			tty.writes(field[I_DATA][field[I_VALUE][0]][1]);
		end
	end
	else if (type = T_DONE) do
		tty.move(x+field[I_X], y+field[I_Y]);
		tty.color(Colorset[C_SELECT]);
		tty.writes(field[I_DATA]);
		k := tty.readc();
		if (k = '\r' \/ k = ttyctl.K_RIGHT) k := Keyset[K_DONE];
		tty.move(x+field[I_X], y+field[I_Y]);
		tty.color(Colorset[C_SOTEXT]);
		tty.writes(field[I_DATA]);
	end
	return k;
 end

 next(items, here) do var start, type;
	start := here;
	while (1) do
		here := here+1;
		if (items[here] = 0) here := 0;
		if (here = start) leave;
		type := items[here][I_TYPE];
		if (type = T_INPUT \/ type = T_MENU \/ type = T_DONE)
			leave;
	end
	return here;
 end

 prev(items, here) do var start, type;
	start := here;
	while (1) do
		here := here-1;
		if (here < 0) do
			for (here=0, 999) if (\items[here]) leave;
			here := here-1;
		end
		if (here = start) leave;
		type := items[here][I_TYPE];
		if (type = T_INPUT \/ type = T_MENU \/ type = T_DONE)
			leave;
	end
	return here;
 end

 public complete(x, y, dx, dy, cs, es, fs, ks, items) do
	var	here, red;
	var	k;

	dx := dx+x;
	dy := dy+y;
	if (dx-x .>= MAXWIDTH) return -1;
	init();
	t.memfill(Spaces, '\s', dx-x);
	Spaces::(dx-x) := 0;
	Itemlist := items;
	Colorset := cs-> cs: Dflcset;
	Ecolrset := es-> es: Dfleset;
	Frameset := fs-> fs: Dflfset;
	Keyset := ks-> ks: Dflkset;
	framebox(x, y, dx, dy);
	red := 1;
	here := items[0][I_TYPE] = T_TEXT-> next(items, 0): 0;
	while (1) do
		if (red) do
			refresh(x, y);
			red := 0;
		end
		k := edit(x, y, dx, dy, items[here]);
		ie (k = Keyset[K_NEXT] \/ k = '\t')
			here := next(items, here);
		else ie (k = Keyset[K_PREV])
			here := prev(items, here);
		else if (k = Keyset[K_DONE] \/ k = ttyctl.K_ESC)
			leave;
	end
	tty.color(0x07);
	fini();
 end

end

do end
