! TIDE -- Persistent structures class
! Copyright (C) 2002 Nils M Holm. All rights reserved.
! See the file LICENSE for conditions of use.

module pstruct();

public class pstruct(t3x, char, string, iostream)

 const MEMBERS = 128;

 object t[t3x], str[string], chr[char];

 public const	FREAD	= IOSTREAM.FREAD,
		FWRITE	= IOSTREAM.FWRITE;

 public const	ERR_IO	= -1,
		ERR_EOF	= -2,
		ERR_FMT	= -3,
		ERR_MEM	= -4;
 const		ERR_EOT = -5;

 object	Ios[iostream];
 var	Alloc;

 public open(file, buf, len, mode)
	return Ios.open(file, buf, len, mode | IOSTREAM.FKILLCR);

 public close() return Ios.close();

 public eof() return Ios.eof();

 ioswrite(s, data, len) 
	return send(s, iostream, write(data, len)) \= len;

 ioswrites(s, text) return ioswrite(s, text, str.length(text));

 ioswrch(s, c) do var buf::5;
	ie (c >= '\s' /\ c <= '~') do
		return send(s, iostream, wrch(c)) = -1;
	end
	else do
		return ioswrites(s, str.format(buf, "\\%2:0x", [(c)]));
	end
 end

 ioswritestr(s, text) do var i;
	i := 0;
	while (text::i) do
		if (ioswrch(s, text::i) = -1) return ERR_IO;
		i := i+1;
	end
 end

 iosrdch(s) do var c, d, digits;
	c := send(s, iostream, rdch());
	if (c < 0) return c;
	if (c = '\\') do
		digits := "0123456789ABCDEF";
		c := send(s, iostream, rdch());
		if (c < 0) return ERR_IO;
		c := str.scan(digits, c);
		if (c < 0) return ERR_IO;
		d := send(s, iostream, rdch());
		if (d < 0) return ERR_IO;
		d := str.scan(digits, d);
		if (d < 0) return ERR_IO;
		return c<<4 | d | 256;
	end
	return c;
 end

 ioseof(s) return send(s, iostream, eof());

 dowrite(s, tmpl, data, pi, pp, indent) do var i, p, q, j, buf::64, r;
	if (tmpl::pi[0] = '[') do
		for (j=0, indent)
			if (ioswrch(s, '\s') = -1)
				return ERR_IO;
	end
	r := 0;
	i := pi[0];
	p := pp[0];
	ie (tmpl::i = 'B') do
		r := r | ioswrites(s, "0b");
		r := r | ioswrites(s, str.numtostr(buf, data[p], 2));
	end
	else ie (tmpl::i = 'C') do
		r := r | ioswrch(s, ''');
		r := r | ioswrch(s, data[p]);
		r := r | ioswrch(s, ''');
	end
	else ie (tmpl::i = 'N') do
		r := r | ioswrites(s, str.numtostr(buf, data[p], -10));
	end
	else ie (tmpl::i = 'S') do
		r := r | ioswrch(s, '"');
		r := r | ioswritestr(s, data[p]);
		r := r | ioswrch(s, '"');
	end
	else ie (tmpl::i = 'X') do
		r := r | ioswrites(s, "0x");
		r := r | ioswrites(s, str.numtostr(buf, data[p], 16));
	end
	else ie (tmpl::i = '[') do
		i := i+1;
		j := i;
		q := 0;
		r := r | ioswrites(s, "[");
		if (tmpl::i = '[')
			r := r | (send(s, iostream, wrch('\n')) = -1);
		while (\r /\ tmpl::i /\ tmpl::i \= ']') do
			if (tmpl::i \= '[' /\ tmpl::i \= '*')
				r := r | ioswrch(s, '\s');
			r := r | dowrite(s, tmpl, data[p], @i, @q, indent+1);
		end
		ie (tmpl::(i-1) = ']' \/ tmpl::(i-1) = '*')
			r := r | ioswrites(s, indent->"]\n": "]");
		else
			r := r | ioswrites(s, indent->" ]\n": " ]");
	end
	else ie (tmpl::i = '*') do
		if (\i) return ERR_FMT;
		if (data[p]) do
			i := i-1;
			if (tmpl::i \= ']') return ERR_FMT;
			while (i /\ tmpl::i \= '[') i := i-1;
			pi[0] := i;
			return 0;
		end
	end
	else do
		return ERR_FMT;
	end
	pi[0] := i+1;
	pp[0] := p+1;
	return r;
 end

 public writeto(s, tmpl, data) do var i, r, p;
	i := 0;
	p := 0;
	r := 0;
	while (\r /\ tmpl::i) do
		r := r | dowrite(s, tmpl, data, @i, @p, 0);
		r := r | (send(s, iostream, wrch('\n')) = -1);
	end
	return r;
 end

 readnum(s, c, pv) do var radix, sign, i, buf::MEMBERS+1;
	sign := 1;
	if (c = '-') do
		c := iosrdch(s);
		sign := -1;
	end
	radix := 10;
	if (c = '0') do
		c := iosrdch(s);
		ie (c < 0) do
			pv[0] := 0;
			return 0;
		end
		else ie (c = 'x' \/ c = 'X') do
			radix := 16;
			c := iosrdch(s);
		end
		else if (c = 'b' \/ c = 'B') do
			radix := 2;
			c := iosrdch(s);
		end
	end
	i := 0;
	while (chr.digit(c) \/ 'A' <= c /\ c <= 'F') do
		if (i >= MEMBERS) return ERR_FMT;
		buf::i := c;
		i := i+1;
		c := iosrdch(s);
	end
	buf::i := 0;
	pv[0] := str.strtonum(buf, radix, 0)*sign;
 end

 readstr(s, pv) do var c, buf::MEMBERS+1, i;
	i := 0;
	while (1) do
		c := iosrdch(s);
		if (c = '"') leave;
		if (i >= MEMBERS) return ERR_FMT;
		buf::i := c & 255;
		i := i+1;
	end
	buf::i := 0;
	i := i+1;
	pv[0] := call Alloc(i);
	if (\pv[0]) return ERR_MEM;
	str.copy(pv[0], buf);
 end

 decl readtbl(2);

 public readfrom(s, palloc, pstat) do var c, v;
	Alloc := palloc;
	chr.init();
	if (pstat) pstat[0] := 0;
	c := iosrdch(s);
	while (1) do
		while (c = '\s' \/ c = '\t' \/ c = '\n' \/ c = '\r')
			c := iosrdch(s);
		if (c \= '#') leave;
		while (\ioseof(s) /\ c \= '\n')
			c := iosrdch(s);
	end
	if (ioseof(s)) do
		if (pstat) pstat[0] := ERR_EOF;
		return ERR_EOF;
	end
	ie (chr.digit(c) \/ c = '-') do
		if (readnum(s, c, @v)) do
			if (pstat) pstat[0] := ERR_FMT;
			return ERR_FMT;
		end
		return v;
	end
	else ie (c = ''') do
		c := iosrdch(s);
		if (c < 0) return ERR_FMT;
		if (iosrdch(s) \= ''') return ERR_FMT;
		return c & 255;
	end
	else ie (c = '"') do
		if (readstr(s, @v)) do
			if (pstat) pstat[0] := ERR_FMT;
			return ERR_FMT;
		end
		return v;
	end
	else ie (c = '[') do
		if (readtbl(s, @v)) do
			if (pstat) pstat[0] := ERR_FMT;
			return ERR_FMT;
		end
		return v;
	end
	else ie (c = ']') do
		if (pstat) pstat[0] := ERR_EOT;
		return ERR_EOT;
	end
	else do
		if (pstat) pstat[0] := ERR_FMT;
		return ERR_FMT;
	end
 end

 readtbl(s, pv) do var buf[MEMBERS+1], p, stat, i, v;
	i := 0;
	while (1) do
		v := self.readfrom(s, Alloc, @stat);
		if (stat = ERR_EOT) leave;
		if (stat) return ERR_FMT;
		if (i >= MEMBERS-1) return ERR_FMT;
		buf[i] := v;
		i := i+1;
	end
	buf[i] := 0;
	i := i+1;
	pv[0] := call Alloc(i*t.bpw());
	if (\pv[0]) return ERR_MEM;
	t.memcopy(pv[0], buf, i*t.bpw());
 end

 public write(tmpl, data) return self.writeto(Ios, tmpl, data);

 public read(pa, pstat) return self.readfrom(Ios, pa, pstat);

end

do end
