! TIDE -- Visual editor class
! Copyright (C) 2000,2001,2002 Nils M Holm. All rights reserved.
! See the file LICENSE for conditions of use.

module vedit();

public class vedit(t3x, string, ttyctl, editor)

 object t[t3x], str[string], tty[ttyctl];
 object	e[editor];

 public struct	SIZES = S_LENGTH, S_TOTAL, S_LINE;

 public struct	CONFIG = C_TABSIZE, C_ASYNC, C_SHOWCTL, C_VBLOCKS,
		C_TEXTCOLOR, C_HIGHCOLOR, C_MARKCOLOR, C_XBASE,
		C_YBASE, C_XMAX, C_YMAX;

 const	MAXWIDTH = 128;
 const	MAXLINE = 256;

 var	Xpos, Ypos, Head, Tail, Offset;
 var	Xbase, Ybase, Xmax, Ymax;
 var	Flatx, Stx;
 var	Spaces::MAXWIDTH+1;

 var	Tabsize;
 var	Async, Delayed;
 var	Showctl;
 var	Vblocks;
 var	Textcolor, Highcolor, Markcolor;

 public reset() do
	e.reset();
	Xpos := 0;
	Ypos := 0;
	Head := 0;
	Stx := 0;
	Offset := 0;
 end

 public init(x, y, width, height, buf, len) do
	if (width > MAXWIDTH) return -1;
	e.init(buf, len);
	tty.init();
	tty.mode(1);
	Xbase := x;
	Ybase := y;
	Xmax := width;
	Ymax := height;
	t.memfill(Spaces, '\s', width);
	Spaces::width := 0;
	Tabsize := 8;
	Async := 1; Delayed := 0;
	Showctl := 1;
	Vblocks := 1;
	Textcolor := ttyctl.F_GREY | ttyctl.B_BLACK;
	Highcolor := ttyctl.F_BLACK | ttyctl.B_GREY;
	Markcolor := ttyctl.F_BLACK | ttyctl.B_GREY;
	tty.color(Textcolor);
	self.reset();
 end

 public fini() do
	tty.color(0x07);
	tty.mode(0);
	tty.fini();
 end

 public _point() return e._point();

 public _line() return e._line();

 public _block() return e._block();

 public _eblock() return e._eblock();

 public _length() return e._length();

 public _char() return e._char();

 public _size(sv) return e._size(sv);

 public _xpos() return Xpos;

 public _ypos() return Ypos;

 flatpos(x) do var buf::MAXLINE+1, rx, i;
	e.copy(e._line(), buf, MAXLINE);
	rx := 0;
	for (i=0, x)
		rx := rx + (buf::i = '\t'->
			Tabsize - (rx+Tabsize) mod Tabsize: 1);
	return rx;
 end

 adjust() do var buf::MAXLINE+1, rx, k, i;
	k := e.copy(e._line(), buf, MAXLINE+1)-1;
	rx := 0;
	for (i=0, k) do
		if (rx >= Stx) leave;
		rx := rx + (buf::i = '\t'->
			Tabsize - (rx+Tabsize) mod Tabsize: 1);
	end
	Xpos := i;
	e.advance(i);
 end

 expand(src, dest, lim) do var i, j, c;
	i := 0;
	j := 0;
	while (j < lim) do
		c := src::i;
		ie (\c) do
			leave;
		end
		else ie (c = '\t') do
			while (1) do
				dest::j := '\s';
				j := j+1;
				if (j mod Tabsize = 0) leave;
			end
		end
		else do
			dest::j := c;
			j := j+1;
		end
		i := i+1;
	end
	if (dest::(j-1) = '\n') j := j-1;
	dest::j := 0;
	return j;
 end

 vprint(s, k, bs, be) do var i, j, c, cc, cf, nc, bc, bf;
	cc := 0;
	j := 0;
	bc := Textcolor;
	bf := 0;
	for (i=0, k) do
		c := s::i;
		if (Vblocks /\ bs \= be) do
			if (i = bs) do nc := Markcolor; bf := 1; end
			if (i = be) do nc := Textcolor; bf := 1; end
		end
		cf := \cc /\ '\s' > c \/ cc /\ '\s' <= c;
		if (cf \/ bf) do
			s::i := 0;
			tty.color(cc->Highcolor: bc);
			tty.writes(@s::j);
			s::i := c;
			j := i;
			if (bf) bc := nc;
			if (cf) cc := \cc;
			bf := 0;
		end
		if (cc) s::i := c+'@' & 127;
	end
	if (i \= j) do
		tty.color(cc->Highcolor: bc);
		tty.writes(@s::j);
	end
 end

 getblockmargs(line, len, x, ps, pe) do var bs, be;
	bs := e._block();
	be := e._eblock();
	ie (bs >= be) do
		bs := 0;
		be := 0;
	end
	else do
		bs := bs > line+len-> MAXLINE:
			bs < line-> 0:
			flatpos(bs-line);
		be := be < line-> 0:
			be > line+len-> MAXLINE:
			flatpos(be-line);
		bs := bs-Offset-x;
		if (bs < 0) bs := 0;
		be := be-Offset-x;
		if (be < 0) be := 0;
	end
	ps[0] := bs;
	pe[0] := be;
 end

 println(x) do
	var	buf::MAXLINE+1, ebuf::MAXLINE+1;
	var	k, i, np, line, len, bs, be;

	line := e._line();
	len := e.copy(line, buf, MAXLINE+1);
	k := expand(buf, ebuf, MAXLINE);
	ebuf::(Xmax+Offset) := 0;
	if (\(Vblocks \/ Showctl)) tty.color(Textcolor);
	ie (k >= Offset) do
		k := str.length(@ebuf::(Offset+x));
		ie (Showctl \/ Vblocks)
			ie (Vblocks) do
				getblockmargs(line, len, x, @bs, @be);
				vprint(@ebuf::(Offset+x), k, bs, be);
			end
			else do
				vprint(@ebuf::(Offset+x), k, 0, 0);
			end
		else
			tty.writes(@ebuf::(Offset+x));
		if (x+k < Xmax) do
			if (Vblocks) tty.color(Textcolor);
			tty.writes(@Spaces::(x+k));
		end
	end
	else do
		tty.writes(Spaces);
	end
 end

 public refresh(first, last) do
	var	i, k, pt, stop;
	var	buf::MAXLINE+1, ebuf::MAXLINE+1;;

	pt := e._point();
	e.goto(Head);
	stop := 0;
	if (first) e.next(first);
	for (i=first, last+1) do
		if (Async /\ tty.query()) do
			Delayed := 1;
			e.goto(pt);
			return;
		end
		tty.move(Xbase, Ybase+i);
		ie (\stop) do
			println(0);
			stop := e.next(1);
		end
		else do
			tty.writes(Spaces);
		end
	end
	Tail := e._line();
	e.goto(pt);
 end

 public display() self.refresh(0, Ymax-1);

 public setup(cv) do var i, p, v, r;
	i := 0;
	r := 0;
	while (cv[i]) do
		p := cv[i][0];
		v := cv[i][1];
		ie (p = C_TABSIZE) Tabsize := v;
		else ie (p = C_ASYNC) Async := v;
		else ie (p = C_SHOWCTL) Showctl := v;
		else ie (p = C_VBLOCKS) Vblocks := v;
		else ie (p = C_TEXTCOLOR) Textcolor := v;
		else ie (p = C_HIGHCOLOR) Highcolor := v;
		else ie (p = C_MARKCOLOR) Markcolor := v;
		else ie (p = C_XBASE) Xbase := v;
		else ie (p = C_YBASE) Ybase := v;
		else ie (p = C_XMAX) Xmax := v;
		else ie (p = C_YMAX) Ymax := v;
		else r := r-1;
		i := i+1;
	end
	t.memfill(Spaces, '\s', Xmax);
	Spaces::Xmax := 0;
	tty.color(Textcolor);
	return r;
 end

 public sync() do var x, r;
	Flatx := flatpos(Xpos);
	r := Offset;
	while (Flatx < Offset) Offset := Offset-Tabsize;
	while (Flatx - Offset >= Xmax) Offset := Offset+Tabsize;
	if (r \= Offset) self.display();
	tty.move(Xbase+Flatx-Offset, Ybase+Ypos);
 end

 public goto(p) do var i, r, rest;
	e.goto(Head);
	for (i=0, Ymax) if (e.next(1)) leave;
	rest := e._line();
	ie (p >= Head /\ p < rest) do
		e.goto(Head);
		r := 0;
	end
	else do
		e.goto(p);
		for (i=0, Ymax/2-1) if (e.prev(1)) leave;
		Head := e._line();
		e.goto(Head);
		ie (Async)
			Delayed := 1;
		else
			self.display();
		r := 1;
	end
	Ypos := e.advance(p-Head);
	Xpos := p - e._line();
	Stx := flatpos(Xpos);
	return r;
 end

 public totop() do
	Head := e._point();
	Ypos := 0;
	ie (Async)
		Delayed := 1;
	else
		self.display();
 end

 public key() do
	if (Delayed /\ \tty.query()) do
		Delayed := 0;
		self.display();
	end
	self.sync();
	return tty.readc();
 end

 public mark() do var eb;
	eb := e._eblock();
	e.mark(e._point());
	if (eb) do
		ie (Async)
			Delayed := 1;
		else
			self.display();
	end
 end

 public setblock() do
	e.setblock(e._point());
	ie (Async)
		Delayed := 1;
	else
		self.display();
 end

 public clearblock() do var eb;
	if (e._eblock() = 0) return;
	e.clearblock();
	ie (Async)
		Delayed := 1;
	else
		self.display();
 end

 _down() do var x, pt;
	if (e.next(1) \= 0) return -1;
	Ypos := Ypos + 1;
	if (Ypos >= Ymax) do
		pt := e._point();
		e.goto(Head);
		e.next(1);
		Head := e._line();
		e.goto(pt);
		ie (Async)
			Delayed := 1;
		else
			self.display();
		Ypos := Ymax-1;
	end
	Xpos := 0;
 end

 _up() do var x, pt;
	if (e.prev(1) \= 0) return -1;
	Ypos := Ypos-1;
	if (Ypos < 0) do
		pt := e._point();
		e.goto(Head);
		e.prev(1);
		Head:= e._line();
		e.goto(pt);
		ie (Async)
			Delayed := 1;
		else
			self.display();
		Ypos := 0;
	end
	Xpos := 0;
 end

 public down() do var r;
	r := _down();
	if (\r) adjust();
	return r;
 end

 public up() do var r;
	r := _up();
	if (\r) adjust();
	return r;
 end

 public right() do var nl;
	nl := e.advance(1);
	if (nl < 0) return -1;
	ie (nl) do
		e.goback(1);
		_down();
		Xpos := 0;
		Stx := 0;
	end
	else do
		Xpos := Xpos+1;
		Stx := flatpos(Xpos);
	end
 end

 public left() do var nl;
	nl := e.goback(1);
	if (nl < 0) return -1;
	ie (nl) do
		e.advance(1);
		_up();
		Xpos := e.eol();
		Stx := flatpos(Xpos);
	end
	else do
		Xpos := Xpos-1;
		Stx := flatpos(Xpos);
	end
 end

 public home() do
	e.home();
	Xpos := 0;
	Stx := 0;
 end

 public eol() do
	Xpos := e.eol();
	Stx := flatpos(Xpos);
 end

 public top() do
	Head := 0;
	e.goto(0);
	Xpos := 0;
	Ypos := 0;
	Stx := 0;
	ie (Async)
		Delayed := 1;
	else
		self.display();
 end

 public bottom() do var pt;
	e.goto(e._length());
	pt := e._line();
	if (e.prev(Ymax-1) \= 0) do
		self.top();
		while (e.next(1) = 0) Ypos := Ypos+1;
		return;
	end
	Head := e._line();
	Ypos := Ymax-1;
	e.goto(pt);
	ie (Async)
		Delayed := 1;
	else
		self.display();
	Xpos := 0;
	Stx := 0;
 end

 public next() do
	e.goto(Head);
	ie (e.next(Ymax) = 0) do
		e.prev(1);
		Head := e._line();
		Xpos := 0;
		Stx := 0;
		Ypos := 0;
		ie (Async)
			Delayed := 1;
		else
			self.display();
	end
	else do
		self.bottom();
	end
 end

 public prev() do
	e.goto(Head);
	ie (e.prev(Ymax-1) = 0) do
		Head := e._line();
		Xpos := 0;
		Stx := 0;
		Ypos := 0;
		ie (Async)
			Delayed := 1;
		else
			self.display();
	end
	else do
		self.top();
	end
 end

 public find(s, rs, re) do var p;
	p := e.find(s, rs, re);
	if (p < 0) return -1;
	self.goto(p);
 end

 public rfind(s, rs, re) do var p;
	p := e.rfind(s, rs, re);
	if (p < 0) return -1;
	self.goto(p);
 end

 public gotoln(n) do
	if (e.gotoln(n)) return -1;
	self.goto(e._point());
 end

 public insert(k) do
	self.clearblock();
	if (e.insertc(k) < 0) return -1;
	ie (k = '\n')
		if (Ypos < Ymax-1) self.refresh(Ypos, Ymax-1);
	else
		println(Flatx-Offset);
	self.right();
 end

 public indent() do var pl::MAXLINE+1, k, c, i;
	self.clearblock();
	if (e.prev(1) \= 0) return -1;
	k := e.copyln(pl, MAXLINE+1)-1;
	e.next(1);
	Xpos := 0;
	Stx := 0;
	for (i=0, k) do
		c := pl::i;
		ie (c = '\s' \/ c = '\t') do
			e.insertc(c);
			self.right();
		end
		else do
			leave;
		end
	end
	tty.move(Xbase, Ybase+Ypos);
	println(0);
 end

 public delete() do
	self.clearblock();
	ie (e.delete(1) > 0) do
		self.refresh(Ypos, Ymax-1);
	end
	else do
		println(flatpos(Xpos)-Offset);
	end
 end

 public backspace() do
	self.clearblock();
	if (self.left()) return -1;
	ie (e.delete(1) > 0) do
		self.refresh(Ypos, Ymax-1);
	end
	else do
		self.sync();
		println(flatpos(Xpos)-Offset);
	end
 end

 public deleteln() do
	self.home();
	if (\e.deleteln()) self.refresh(Ypos, Ymax-1);
 end

 public readblock(fname) do var r;
	r := e.readblock(fname);
	if (r >= 0) self.display();
	return r;
 end

 public writeblock(fname) return e.writeblock(fname);

 public appendblock(fname) return e.appendblock(fname);

 public deleteblock() do
	if (e._eblock() <= e._block()) return -1;
	self.goto(e._block());
	e.deleteblock();
	ie (Async)
		Delayed := 1;
	else
		self.display();
 end

 public copyblock() do var r;
	r := e.copyblock();
	if (\r) do
		ie (Async)
			Delayed := 1;
		else
			self.display();
	end
	return r;
 end

 public moveblock() do var r;
	r := e.moveblock();
	if (\r) do
		if (\self.goto(e._block())) do
			ie (Async)
				Delayed := 1;
			else
				self.display();
		end
	end
	return r;
 end

 public findmatch() do var p;
	p := e.findmatch();
	if (p < 0) return -1;
	self.goto(p);
 end

 public showmatch(on) do var p, q, x;
	p := e.findmatch();
	if (p < Head \/ p >= Tail) return;
	q := e._point();
	e.goto(p);
	x := flatpos(e._point() - e._line());
	if (x < Offset \/ x >= Offset + Xmax) do
		e.goto(q);
		return;
	end
	self.goto(p);
	self.sync();
	tty.color(on-> Highcolor:
		(p>=e._block() /\ p<e._eblock())-> Markcolor: Textcolor);
	tty.writec(e._char());
	self.goto(q);
 end

 public marklist() do
	e.marklist();
	ie (Async)
		Delayed := 1;
	else
		self.display();
 end

 public load(name) do var r;
	r := e.load(name);
	self.top();
	return r;
 end

 public load_dos(name) do var r;
	r := e.load_dos(name);
	self.top();
	return r;
 end

 public save(name) return e.save(name);

 public save_dos(name) return e.save_dos(name);

 public copyln(buf, len) return e.copyln(buf, len);

 public nv_find(s, rs, re) return e.find(s, rs, re);

 public nv_goto(p) return e.goto(p);

 public nv_delete(n) return e.delete(n);

 public nv_insert(s) return e.inserts(s);

 public ttymode(m) tty.mode(m);

end

do end
